@@ -2,6 +2,9 @@ from fractions import Fraction
 import pickle
 import copy
 from math import log, exp
+import pandas as pd
+import matplotlib.pyplot as plt
+import numpy as np
 
 general = {
     "det()": ["matrix M", "Returns the determinant of the given matrix.", "det"],
@@ -46,11 +49,9 @@ general = {
     "rInverse()": ["matrix M", "Computes the inverse of a matrix by putting an augmented matrix into RREF using Gauss-Jordan elimination, and returns this inverse."],
     "elimSystem()": ["matrix A, vector x, vector b, num r=1000", "Computes the solutions to the system Ax=b, using Gauss-Jordan elimination. A must be a square matrix of coefficients, x and b should be column vectors in string and numerical formats respectively. r is the rational-denominator limit. The output will be a list, including both the solution matrix and a series of solution strings.", "elimSystem"],
     "isMatrix()": ["obj A", "Returns True if the passed object is a matrix, otherwise returns False."],
-    "nthRegression()": ["num r=0, matrix A=['empty'], num n=1, num rcomp=1", "Computes an nth-degree regression by the least squares method and returns the best fit polynomial. Printing output restricted if r==1. Matrix input and use of n (degree preset) only intended for use by other dependent regression functions. R^2 coefficient of determination will be computed if rcomp==1."],
-    "pwrRegression()": ["num r=0, num rcomp=1", "Computes a power regression by a logarithmic-linear least squares method and returns the best fitting function y=ax^b. Printing output restricted if r==1. R^2 coefficient of determination will be computed if rcomp==1."],
-    "expRegression()": ["num r=0, num rcomp=1", "Computes an exponential regression by a logarithmic-linear least squares method and returns the best fitting function y=ab^x. Printing output restricted if r==1. R^2 coefficient of determination will be computed if rcomp==1."]
-}
-
+    "regression()": ["matrix M, str model='lin', str degree='2', str type='ls', num rcomp=1, num restrict=0", "Uses a regression model to compute a best-fitting curve to a given dataset. Includes linear, exponential, power, polynomial (nth-degree), and conic regression. All but polynomial and conic can be given a regression type, either least-squares (ls) or Theil-Sen (robust). Unless rcomp==0, an R^2 coefficient of determination value and an R_a absolute-residual coefficient will be calculated. Set restrict==0 to print at the end of calculations. Output is a list [[coeffs],function,[R^2,R_a]], where function = output[1] = a function formatted to work with the matplotlib graphing utilities (except for conics!), and the output[2] list will not exist if rcomp!=1."],
+    "regPlot()": ["str src, str filetype='CSV', str model='lin', str type='ls', int deg=2, str xlab='X Axis', str ylab='Y Axis', int prec=10", "Takes an input data file (src) using pandas and plots both the data and a curve of best fit. Type should be the file type, such as csv. xlab and ylab label plot axes. prec sets the precision - higher numbers allow for higher precision plotting, but longer computation time."]
+    }
 
 def det(mat):
     matrix = copy.deepcopy(mat)
@@ -336,8 +337,7 @@ def ratMPrint(mat, s=1, l=0):
         for a in range(0, s):
             spaces += " "
         for j in range(0, n):
-            str += ("%s%s") % (spaces, Fraction(matrix[i]
-                                                [j]).limit_denominator())
+            str += ("%s%s") % (spaces, Fraction(matrix[i][j]).limit_denominator())
         print(str)
         for k in range(0, l):
             print(" ")
@@ -485,7 +485,7 @@ def help(rep=0):
 
 def modulehelp(rep=0):
     if(rep == 0):
-        print("This is the Matrix Module V0.0.11, a package created by Elliott Walker. Type 'funcs' for a list of available functions, 'details' for a more detailed list including parameters, or 'exit' to exit this help function.")
+        print("This is the Matrix Module V0.1.4, a package created by Elliott Walker. Type 'funcs' for a list of available functions, 'details' for a more detailed list including parameters, or 'exit' to exit this help function.")
     else:
         print("Available commands: 'funcs', 'details', 'exit'")
     inp = input()
@@ -546,6 +546,7 @@ def remove(key):
         del index[key]
     return
 
+
 def addIndex(matrix, key):
     if (matrix.isMatrix == True):
         if (key in index.keys()):
@@ -556,6 +557,7 @@ def addIndex(matrix, key):
         print("Program failure: Object passed is not a matrix.")
     return
 
+
 def listMatrices(rat=1):
     print("All stored matrices:")
     for key in index.keys():
@@ -569,6 +571,7 @@ def listMatrices(rat=1):
     print("Access a stored matrix using MatrixModule.index['key'].")
     return
 
+
 def saveMatrices(filename="matrices.txt"):
     # Saves the matrices index to a file using the pickle dump method.
     file=open(filename, "wb")
@@ -707,6 +710,7 @@ def inputSystem(rat=1000, suppress=0, readback=1):
         input("Press return to exit.")
     return(output)
 
+
 def isMatrix(obj):
     if (type(obj)==list):
         for i in range(len(obj)):
@@ -716,179 +720,157 @@ def isMatrix(obj):
     else:
         return False
 
-def nthRegression(restrict=0,inMatrix=["empty"],ndeg=1,rcomp=1):
-    # Performs least-squares regression in order to find the best fitting nth-degree polynomial. For n > 6 the computation may take a long time / not work, as computational expense goes up exponentially with increasing n (inverse of (n+1)x(n+1) matrix must be taken...)
-    npoints = 1
-    if(inMatrix==["empty"]):
-        print("What degree regression would you like to perform?")
-        ndeg = int(input())
-        print("Would you like to use Matrix input (1) or Manual input (2)?")
-        inmode = input()
-        dataMt = []
-        if (int(inmode) == 1):
-            # I don't have anything to handle input matrices that are too small at the moment.
-            print("Enter index of the dataset matrix to input. For example, to access the matrix at location index['matA'], simply type 'matA' (no quotes).")
-            indx = input()
-            dataMt = index[str(indx)]
-        else:
-            print("How many datapoints would you like to input? You must have at least %d datapoints. When entering datapoints use the format x,y."%(ndeg))
-            npoints = int(input())
-            if(npoints < ndeg):
-                print("Not enough datapoints. Failed.")
-                return
-            for i in range(npoints):
-                print("Enter datapoint at index %d:"%(i))
-                ptIn = input()
-                ptArr = ptIn.split(",")
-                if(len(ptArr) != 2):
-                    print("Program failure: Improper formatting of input.")
-                    return
-                dataMt.append([float(ptArr[0]),float(ptArr[1])])
-    else:
-        dataMt = inMatrix
-        npoints = len(dataMt)
-    xset = []
-    yset = []
-    for i in range(npoints):
-        yset.append([dataMt[i][1]])
-        rowMat = []
-        xcur = dataMt[i][0]
-        for j in range(ndeg+1):
-            rowMat.append(xcur**j)
-        xset.append(rowMat)
-    coeffs = multiplyMatrices(multiplyMatrices(inverse(multiplyMatrices(transpose(xset),xset)),transpose(xset)),yset)
-    outpoly = "y = "
-    for i in range(ndeg+1):
-        outpoly += ("%sx^%d" % (Fraction(coeffs[i][0]).limit_denominator(), i)) # Readable format y = ax^0 + bx^1 + cx^2 + ...
-        if(i < ndeg):
-            outpoly += " + "
-    if(restrict != 1):
-        print("outpoly")
-    R2 = 0
-    if(rcomp == 1):
-        def polyFunction(x):
-            y = 0
-            for i in range(ndeg+1):
-                y += coeffs[i][0]*(x**i)
-            return y
-        # Compute R^2 value (coefficient of determination)
-        # Begin by computing RSS (Residual Sum of Squares)
-        RSS = 0
-        for i in range(len(dataMt)):
-            RSS += (dataMt[i][1]-polyFunction(dataMt[i][0]))**2 # Sum of squares of residuals (errors between dataset values and model prediction)
-        # Now compute TSS (Total Sum of Squares)
-        TSS = 0
-        mu = 0 # Mean. Will sum all values (requiring starting value of 0) and then divide by number of values.
-        for i in range(len(dataMt)):
-            mu += dataMt[i][1]
-        mu /= len(dataMt)
-        for i in range(len(dataMt)):
-            TSS += (dataMt[i][1]-mu)**2 # Total sum of squares, which is proportional to variance. Sum of squares of differences between y-values and mean of y-values.
-        R2 = 1 - RSS/TSS # General definition for R^2 given RSS and TSS, see Coefficient of Determination on Wikipedia.
-        print("R^2 = %s" % (R2))
-    return [coeffs,outpoly,R2]
-    
-def pwrRegression(restrict=0,rcomp=1):
-    # Computes the closest power-regression (form of y=ax^b) fit to a dataset using logarithmic least-squares
-    print("Would you like to use Matrix input (1) or Manual input (2)?")
-    inmode = input()
-    lnMt = []
-    if (int(inmode) == 1):
-        # I don't have anything to handle input matrices that are too small at the moment.
-        print("Enter index of the dataset matrix to input. For example, to access the matrix at location index['matA'], simply type 'matA' (no quotes).")
-        indx = input()
-        lnMt = index[str(indx)]
-    else:
-        print("How many datapoints would you like to input? You must have at least 3 datapoints. When entering datapoints use the format x,y.")
-        npoints = int(input())
-        if(npoints < 3):
-            print("Not enough datapoints. Failed.")
-            return
-        for i in range(npoints):
-            print("Enter datapoint at index %d:"%(i))
-            ptIn = input()
-            ptArr = ptIn.split(",")
-            if(len(ptArr) != 2):
-                print("Program failure: Improper formatting of input.")
-                return
-            lnMt.append([log(float(ptArr[0])),log(float(ptArr[1]))]) # Converts coordinates to (log(x),log(y)) to exploit fact that log(ax^b) = log(a) + b*log(x), which is linear.
-    lnMt_reg = nthRegression(1,lnMt,1,0)[0] # Performs linear regression (n=1) on logarithmic dataset
-    lnMt_reg[0][0] = exp(lnMt_reg[0][0]) # Accounts for fact that "y-intercept" value is still in natural logarithm form by performing exponentiation
-    outeq = "y = %sx^(%s)" % (Fraction(lnMt_reg[0][0]).limit_denominator(),Fraction(lnMt_reg[1][0]).limit_denominator()) # Readable format y = ax^(b)
-    if(restrict != 1):
-        print(outeq)
-    R2 = 0
-    if(rcomp == 1):
-        a = lnMt_reg[0][0]
-        b = lnMt_reg[1][0]
-        def pwrFunction(x):
-            return a*(x**b)
-        # Compute R^2 value (coefficient of determination)
-        # Begin by computing RSS (Residual Sum of Squares)
-        RSS = 0
-        for i in range(len(lnMt)):
-            RSS += (exp(lnMt[i][1])-pwrFunction(exp(lnMt[i][0])))**2 # Sum of squares of residuals (errors between dataset values and model prediction)
-        # Now compute TSS (Total Sum of Squares)
-        TSS = 0
-        mu = 0 # Mean. Will sum all values (requiring starting value of 0) and then divide by number of values.
-        for i in range(len(lnMt)):
-            mu += exp(lnMt[i][1])
-        mu /= len(lnMt)
-        for i in range(len(lnMt)):
-            TSS += (exp(lnMt[i][1])-mu)**2 # Total sum of squares, which is proportional to variance. Sum of squares of differences between y-values and mean of y-values.
-        R2 = 1 - RSS/TSS # General definition for R^2 given RSS and TSS, see Coefficient of Determination on Wikipedia.
-        print("R^2 = %s" % (R2))
-    return [lnMt_reg,outeq,R2]
 
-def expRegression(restrict=0,rcomp=1):
-    # Computes the closest exponential-regression (form of y=ab^x) fit to a dataset using logarithmic least-squares
-    print("Would you like to use Matrix input (1) or Manual input (2)?")
-    inmode = input()
-    lnMt = []
-    if (int(inmode) == 1):
-        # I don't have anything to handle input matrices that are too small at the moment.
-        print("Enter index key of the dataset matrix to input.")
-        indx = input()
-        lnMt = index[str(indx)] # At the current stage of development, this needs to already be in (x,log(y)) format.
-    else:
-        print("How many datapoints would you like to input? You must have at least 3 datapoints. When entering datapoints use the format x,y.")
-        npoints = int(input())
-        if(npoints < 3):
-            print("Not enough datapoints. Failed.")
-            return
-        for i in range(npoints):
-            print("Enter datapoint at index %d:"%(i))
-            ptIn = input()
-            ptArr = ptIn.split(",")
-            if(len(ptArr) != 2):
-                print("Program failure: Improper formatting of input.")
-                return
-            lnMt.append([float(ptArr[0]),log(float(ptArr[1]))]) # Converts coordinates to (x,log(y)) to exploit fact that log(ab^x) = log(a) + x*log(b), which is linear.
-    lnMt_reg = nthRegression(1,lnMt,1,0)[0] # Performs linear regression (n=1) on logarithmic dataset
-    lnMt_reg[0][0] = exp(lnMt_reg[0][0]) # Account for a-value still being in natural logarithm form by exponentiating
-    lnMt_reg[1][0] = exp(lnMt_reg[1][0]) # Account for b-value logarithm form, as above
-    outeq = "y = %s*%s^x" % (Fraction(lnMt_reg[0][0]).limit_denominator(),Fraction(lnMt_reg[1][0]).limit_denominator()) # Readable format y = ax^(b)
+def regression(inmat,model="lin",degree=2,type="ls",rcomp=1, restrict=1):
+    # Combine the many different types of regression into one separate function (no dependency) that requires matrix input. This will be more compatible with raw datasets, and have a generalized output compatible with matplotlib.
+    # Note: do NOT use the Fraction().limit_denominator() function to create limited-precision rational coefficients. This can cause zero-rounding error.
+    output, A, b = [], [], [] # Initialize output matrix and intermediate matrices A and b.
+    m = len(inmat)
+    func = lambda x: x # Initialize a function. This will be redefined at the end of each regression model.
+    eq = ""
+    if (model == "lin" or model == "linear"): # y=mx+b linear regression
+        slope, intercept = 0, 0
+        if(type=="leastsquares" or type=="ls"): # Least squares regression - based in matrix multiplication
+            for i in range(m):
+                A.append([1,inmat[i][0]]) # Append [1,x] rows to matrix A
+                b.append([inmat[i][1]]) # Append all y values as independent rows in matrified vector b
+            coeffs = multiplyMatrices(multiplyMatrices(inverse(multiplyMatrices(transpose(A),A)),transpose(A)),b) # Perform multiplication. Based in least-squares projection: A x = b --> A^T A x_h = A^T b --> x_h = (A^T A)^-1 A^T b
+            slope = coeffs[1][0]
+            intercept = coeffs[0][0]
+        elif(type=="robust" or type=="theilsen"): # Theil-Sen estimator regression - based in medians. Very high breakdown point, making it great for large and fairly noisy datasets.
+            for i in range(m):
+                for j in range(m):
+                    if(i < j): # For all possible pairs of different points... (using < to make sure pairs are not double-counted)
+                        if(inmat[i][0] != inmat[j][0]): # where the slope is not undefined...
+                            A.append((inmat[i][1]-inmat[j][1])/(inmat[i][0]-inmat[j][0])) # Append the slope of these points into the matrix A
+            A.sort() # Sort the list of slopes to find the median slope between points.
+            slope = A[int(round(len(A)/2))] # Take median
+            for i in range(m):
+                b.append(inmat[i][1] - (slope*inmat[i][0])) # Find the median intercept b provided by iterating through all datapoints with the equation b=y-ax
+            b.sort()
+            intercept = b[int(round(len(b)/2))] # Take median
+        func = lambda x : slope*x+intercept
+        eq = "%s*x+%s" % (slope, intercept) 
+        output.append([slope,intercept])        
+    elif (model == "pwr" or model == "power"): # Power function regression, in the form of y=ax^p. Exploits logarithm properties in order to use a linear regression technique (either least-squares or Theil-Sen) to form a power-function fit.
+        a, p = 0, 0
+        if(type=="leastsquares" or type=="ls"):
+            for i in range(m):  # Take all datapoint inputs (log(x),log(y))
+                A.append([1,log(inmat[i][0])])
+                b.append([log(inmat[i][1])])
+            coeffs = multiplyMatrices(multiplyMatrices(inverse(multiplyMatrices(transpose(A),A)),transpose(A)),b)
+            a = exp(coeffs[0][0]) # Need to exponentiate a, because it is still in its logarithmic form.
+            p = coeffs[1][0]
+        elif(type=="robust" or type=="theilsen"):
+            for i in range(m):
+                for j in range(m):
+                    if(i < j):
+                        if(inmat[i][0] != inmat[j][0]):
+                            A.append((log(inmat[i][1])-log(inmat[j][1]))/(log(inmat[i][0])-log(inmat[j][0])))
+            A.sort()
+            p = A[int(round(len(A)/2))]
+            for i in range(m):
+                b.append(log(inmat[i][1]) - p*log(inmat[i][0]))
+            b.sort()
+            a = exp(b[int(round(len(b)/2))])
+        func = lambda x : a * (x**p)
+        eq = "%s*x**(%s)" % (a,p)
+        output.append([a,p])
+    elif (model == "nth" or model == "ndegree" or model == "poly" or model == "polynomial" or model == "quadratic"): # Polynomial regression to the nth degree. Doesn't work very well if n is greater than 6 or 7, as the matrix equation gets exponentially more expensive to compute.
+        if model=="quadratic": degree=2 # Default to 2nd degree regresion if quadratic is requested.
+        for i in range(m):
+            rowapp = []
+            for j in range(degree+1):
+                rowapp.append(inmat[i][0]**j) # Rows of form 1, x, x^2, ..., x^n where n is the degree
+            A.append(rowapp)
+            b.append([inmat[i][1]])
+        coeffs = multiplyMatrices(multiplyMatrices(inverse(multiplyMatrices(transpose(A),A)),transpose(A)),b)
+        eq = "%s" % (coeffs[0][0])
+        for i in range(degree):
+            eq += ("+%s*x**%s" % (coeffs[i+1][0],i+1))
+        def func(x): # The function for r-computation can't be trivially defined with lambda anymore because an indeterminate number of terms need to be summed.
+            out = 0
+            for i in range(degree+1):
+                out += coeffs[i][0]*(x**i)
+            return out
+        output.append(coeffs)
+    elif (model == "exp" or model == "exponential"): # Exponential function regression, in the form of y=ap^x. Exploits logarithm properties in order to use a linear regression technique (either least-squares or Theil-Sen) to form a power-function fit.
+        a, p = 0, 0
+        if(type=="leastsquares" or type=="ls"):
+            for i in range(m):  # Take all datapoint inputs (x,log(y))
+                A.append([1,inmat[i][0]])
+                b.append([log(inmat[i][1])])
+            coeffs = multiplyMatrices(multiplyMatrices(inverse(multiplyMatrices(transpose(A),A)),transpose(A)),b)
+            a = exp(coeffs[0][0]) # Need to exponentiate both a and p, because they are both in logarithmic form here.
+            p = exp(coeffs[1][0])
+        elif(type=="robust" or type=="theilsen"):
+            for i in range(m):
+                for j in range(m):
+                    if(i < j):
+                        if(inmat[i][0] != inmat[j][0]):
+                            A.append((log(inmat[i][1])-log(inmat[j][1]))/(inmat[i][0]-inmat[j][0])) # Take inputs (x,log(y)) for the Theil-Sen slope estimation
+            A.sort()
+            p = exp(A[int(round(len(A)/2))])
+            for i in range(m):
+                b.append(log(inmat[i][1]) - p*log(inmat[i][0]))
+            b.sort()
+            a = exp(b[int(round(len(b)/2))])
+        func = lambda x : a * (p**x)
+        eq = "%s*(%s)**x" % (a,p)
+        output.append([a,p])
+    elif (model == "con" or model == "conic"): # General conic form is Ax^2 + Bxy + Cy^2 + Dx + Ey + F = 0.
+        rcomp=0 # Can't compute R^2 or R_a by normal method using this
+        for i in range(m):
+            A.append([inmat[i][0]**2,inmat[i][0]*inmat[i][1],inmat[i][1]**2,inmat[i][0],inmat[i][1]])
+            b.append([1]) # I can't set the LHS equal to 0, so instead I set F = -1 and obtain Ax^2 + Bxy + Cy^2 + Dx +Ey = 1. This gives a nonzero b-vector to multiply by.
+        coeffs = multiplyMatrices(multiplyMatrices(inverse(multiplyMatrices(transpose(A),A)),transpose(A)),b)
+        eq = "%s*x**2+%s*xy+%s*y**2+%s*x+%s*y+%s=0" % (coeffs[0][0],coeffs[1][0],coeffs[2][0],coeffs[3][0],coeffs[4][0],-1)
+        func = lambda x : "NaF: Conic" # The conic is defined by an implicit equation of x and y, not by an explicit function.
+        coeffs.append([-1])
+        output.append(coeffs)  
+    output.append(eq)
+    if(rcomp==1):
+        RSS,RSA,TSS,TSA,mu = 0,0,0,0,0
+        for i in range(m):
+            mu += inmat[i][1] # Sum total for mean (mu) computation
+            RSS += (inmat[i][1] - func(inmat[i][0]))**2 # Residual sum of squared errors
+            RSA += abs(inmat[i][1] - func(inmat[i][0])) # Residual sum of absolute errors
+        mu/=m
+        for i in range(m):
+            TSS += (inmat[i][1] - mu)**2 # Total sum of squares (of distances from mean)
+            TSA += abs(inmat[i][1] - mu) # Total sum of absolute distances from mean
+        R2 = 1-RSS/TSS # R^2 coefficient of determination
+        RA = 1-RSA/TSA # R_a residual absolute error coefficient
+        if(restrict != 1):
+            print("R^2 = %s, R_a = %s" % (R2, RA))
+        output.append([R2,RA])
     if(restrict != 1):
-        print(outeq)
-    R2 = 0
-    if(rcomp == 1):
-        a = lnMt_reg[0][0]
-        b = lnMt_reg[1][0]
-        def expFunction(x):
-            return a*(b**x)
-        # Compute R^2 value (coefficient of determination)
-        # Begin by computing RSS (Residual Sum of Squares)
-        RSS = 0
-        for i in range(len(lnMt)):
-            RSS += (exp(lnMt[i][1])-expFunction(lnMt[i][0]))**2 # Sum of squares of residuals (errors between dataset values and model prediction)
-        # Now compute TSS (Total Sum of Squares)
-        TSS = 0
-        mu = 0 # Mean. Will sum all values (requiring starting value of 0) and then divide by number of values.
-        for i in range(len(lnMt)):
-            mu += exp(lnMt[i][1])
-        mu /= len(lnMt)
-        for i in range(len(lnMt)):
-            TSS += (exp(lnMt[i][1])-mu)**2 # Total sum of squares, which is proportional to variance. Sum of squares of differences between y-values and mean of y-values.
-        R2 = 1 - RSS/TSS # General definition for R^2 given RSS and TSS, see Coefficient of Determination on Wikipedia.
-        print("R^2 = %s" % (R2))
-    return [lnMt_reg,outeq,R2]
\ No newline at end of file
+        if(model != "conic" and model != "con"):
+            print("y=" + eq)
+        else:           
+            print(eq)
+    return output
+
+
+def regPlot(src,filetype="CSV",model="lin",type="ls",deg=2,xlab="X Axis",ylab="Y Axis",prec=10):
+    df=[]
+    if filetype in ["CSV", "csv", ".csv"]: df = pd.read_csv(src,names=[xlab,ylab])
+    if filetype in ["json", "JSON"]: df = pd.read_json(src,names=[xlab,ylab])
+    if filetype in ["excel", "Excel", "xlsx", "XLSX"]: df = pd.read_excel(src,names=[xlab,ylab])
+    if filetype in ["pickle", "Pickle"]: df = pd.read_pickle(src,names=[xlab,ylab])
+    arr = np.array(df)
+    minx,maxx=arr[0][0],arr[0][0]
+    for i in range(len(arr)):
+        if (arr[i][0] > maxx): maxx = arr[i][0]
+        if (arr[i][0] < minx): minx = arr[i][0]
+    rangex = maxx - minx
+    x=np.linspace(int(minx-(rangex/100)),int(maxx+(rangex/100)),int(rangex*prec))
+    fig,ax=plt.subplots()
+    plt.xlabel(xlab)
+    plt.ylabel(ylab)
+    y=eval(regression(arr,model=model,type=type,degree=deg,rcomp=0)[1])
+    plt.plot(x,y,'r')
+    xd,yd=arr.T
+    plt.scatter(xd,yd,s=5)
+    plt.show()
+    return
\ No newline at end of file
